local top_screens = set(
  'UnpackScreen',
  'TacticalScreen',
  'PlanetTestScreen',
)

function _G.ensure_galaxy_screen()

  local current_screen = lua_screen_name()

  if current_screen=='GalaxyMap'
    return
  elseif current_screen=='StartScreen'
    push_screen 'GalaxyMap'
  elseif top_screens[current_screen]
    close_screen()
  else
    errorf('unknown screen name %s',current_screen)
  end
  GalaxyMap.completed_attack_id=nil
  print('cleared it!')

end

_ENV = create_file_env(_ENV)

myskin = myskin or GalaxySkin1

CallbackHandler(_ENV)

DebuggingUI(_ENV)

function set_skin(skin)
  myskin = gui[skin]
  if not myskin
    trace_error(format("%q is not a valid skin name",tostring(skin)))
  end
  reload_ui()
end

function reload_ui()
  reload_script([[GUI\~GalaxyMap\GalaxyMap.lua]])
end

depends("GalaxySkin1")
depends 'SystemGlyphs'

StandardFonts(_ENV)

layout = myskin.GalaxyLayout(_ENV)
--fonts = myskin.GalaxyFonts(_ENV)
top_bar = myskin.TopBar(_ENV)
bottom_bar = myskin.BottomBar(_ENV)
action_colors = myskin.ActionColors(_ENV)



local execute_draw_requests, purple_cursor in _ENV



EventDispatcher(_ENV)

ToolTipText(_ENV)
LazyHoverText(_ENV)
ImprovementDrawer(_ENV)

SplitScreenPanes(_ENV)

EnhancedReportPane(_ENV)

EdgePanning(_ENV)

GalaxyActionLinks(_ENV)

planet_bar = myskin.PlanetBar(_ENV)

TopBarContents(_ENV)
RacesPane(_ENV)
TechPane(_ENV)
MainMenu(_ENV)



MapElement(_ENV)

StrategicZoom(_ENV)

BattleListPane(_ENV)
FleetListPane(_ENV)
PlanetListPane(_ENV)
PlanetPopPane(_ENV)



ReportIcons(_ENV)
ReportNotificationIcon(_ENV)

DrawStarCenters(_ENV)
DrawStarOrbits(_ENV)
DrawStarNames(_ENV)
Starfield(_ENV)
RangeCircles(_ENV)
TransportClicks(_ENV)
PlanetRow(_ENV)
StarNameEdit(_ENV)
FleetRoutes(_ENV)
FleetIcons(_ENV)
ShipHoverIcons(_ENV)
FleetShips(_ENV)
TradingShips(_ENV)
--EmpireDrawer(_ENV)

PlanetPane(_ENV)
PlanetPopHovers(_ENV)
PlanetPopInfoBoxes(_ENV)
ProductionQueue(_ENV)
LeaderReport(_ENV)
BattleIcons(_ENV)


ChooseProduction(_ENV)

BattleDraw(_ENV)
OrbitReports(_ENV)
ColonyReports(_ENV)
MoraleReports(_ENV)
TechReports(_ENV)
InfoReports(_ENV)
SystemEncounterReport(_ENV)
GenericEnhancedDialog(_ENV)
ScanReports(_ENV)

AnnounceReports(_ENV)
DiploReports(_ENV)
ShipCaptureReports(_ENV)
CouncilMeetingReports(_ENV)

DesignerPane(_ENV)
ChooseDesign(_ENV)

OptionsPane(_ENV)
EventTriggerPane(_ENV)
SoundTestPane(_ENV)
ResolutionTestPane(_ENV)

StandardPanes(_ENV)
ShipMidscreenPane(_ENV)

BattleSummary(_ENV)
NextTurn(_ENV)

GroundCombatPane(_ENV)
EmbarkPane(_ENV)

DiploPane(_ENV)
FactionBar(_ENV) FactionList(_ENV) FactionDetails(_ENV) 
FactionMapElement(_ENV)
SmallDiploElement(_ENV) MercenaryExchangeElement(_ENV)
ThroneRoomScene(_ENV)
TradingOverlay(_ENV)
DiscussionOverlay(_ENV)
CouncilPane(_ENV)

--TradingPane(_ENV)

function set_desc_text(key)
	local set = rawget(fullscreen_pane?,'set_desc_text')
	if set set(key) end
end


function open_new_design_screen(mode)
  open_fullscreen_pane( choose_design_pane, mode or 'Edit')
end

function open_designer( ...)
  open_fullscreen_pane(designer_pane,...)
end

--open_designer = partial_application(push_fullscreen_pane,designer_pane)

--function set_fullscreen (t)
--	return partial_application(open_fullscreen_pane,t)
--end

function stars_covered()
  return (fullscreen_pane~=split_screen_panes) or midscreen_pane
end

function title_covered()
  return midscreen_pane and (not midscreen_pane.dont_block_bottom_titles)
end

local check_order_callbacks in _ENV

local get_undo_text =  weak_memoizeNN | function(fun,arg)
  local can_undo=undo_string()

  if not can_undo
    return render_hover_text("undo_text", '(nothing to undo)')
  end

  if fun==undo_order
    local str = format('<b>undo action:</b> %s', can_undo)
    str..=vskip_newline(3) .. 'hold <b>SHIFT</b> to instead jump back 1 turn.'
    return render_hover_text('undo1', shorthand_format | str)
  elseif arg==nil
    local str = format('<b>undo last turn</b>')
    str..=vskip_newline(3) .. 'hold <b>SHIFT+CTRL</b> to instead jump back 10 turns.'
    return render_hover_text('undo2', shorthand_format | str)
  end

  assert(fun==jump_to_prev_year)

  return render_hover_text('undo10',|) shorthand_format "<b>undo the last 10 turns</b>"
end

local function undo_click(el)
  if undo_string()
    call_with_args(undo_semantic())
  end
end

local draw_undo = top_hover_fun | function ()
  local undo_text,undo_arrow,undo_tip in top_bar

  local can_undo=undo_string()

  local undo_el = insert(element(undo_arrow(can_undo)),
    'on_hover', can_undo and hand_cursor or grey_cursor,
    'on_click',undo_click,
    sound=can_undo and 'WIDGET_INCREMENT_DOWN' or 'INVALID_ACTION'  )

  FLAG_call('hcentered_tooltip',
    top_hover,top_bar.undox, get_undo_text( undo_semantic() ))
end

local get_redo_text =  weak_memoizeNN | function(fun,arg)
  local can_redo=redo_string()

  if not can_redo
    return render_hover_text('redo_text','(nothing to redo)')
  end

  if fun==redo_order
    local str = format('<b>redo action:</b> %s', can_redo)
    str..=vskip_newline(3) .. 'hold <b>SHIFT</b> to instead jump forward 1 turn.'
    return render_hover_text('redo1', shorthand_format | str)
  elseif arg==nil
    local str = format('<b>jump forward to the next turn</b>')
    str..=vskip_newline(3) .. 'hold <b>SHIFT+CTRL</b> to instead jump forward 10 turns.'
    return render_hover_text('redo2', shorthand_format | str)
  end


  assert(fun==jump_to_next_year)

  return render_hover_text('redo10',|) shorthand_format "<b>jump forward 10 turns</b>"
end

local function redo_click(el)
  if redo_string()
    call_with_args(redo_semantic())
  end
end

local draw_redo= top_hover_fun | function ()
  local redo_text,redo_arrow,undo_tip in top_bar

  local can_redo=redo_string()

  insert(element(redo_arrow(can_redo)),
    'on_hover', can_redo and hand_cursor or grey_cursor,
    'on_click', redo_click,
    sound=can_redo and 'WIDGET_INCREMENT_UP' or 'INVALID_ACTION' )

  FLAG_call('hcentered_tooltip',
    top_hover,top_bar.undox, get_redo_text( redo_semantic() ))

end






function draw_undo_redo()

  draw_zoom_center()
  draw_zoom_out()
  draw_zoom_in()

  if not gui.hide_undo_redo
    draw_undo()
    draw_redo()
  end

end

local function toggle_active_ai()
  local player=gui_player()
  if active_ai(player)
    unset_active_ai(player)
  else
    set_active_ai(player)
  end
end



--local function open_races_pane()
--  open_fullscreen_pane | races_pane
--end

function draw_top_frames()
  top_bar.draw()

  insert( top_bar.end_year_element(),
    'on_click', end_year_click,
    'on_hover', rollover_hand_cursor)

  if options.ai_dev

    insert(element(top_bar.empire_AI(active_ai(gui_player()), gui_player().name )),
      'on_click',toggle_active_ai,
      'on_hover',hand_cursor)

  end

end

local function draw_top_bar()

  FLAG_call('gui_display',SCOPE_call, empire=gui_player(), draw_resources)



  draw_undo_redo()
  top_bar.year_text("Stardate: %d",galaxy.year or 0)


end

export(..draw_top_bar,_ENV)




local function drawtest2()

  call_with_fragment([[Lua State\Drawers\Effects\clouds]],glWrap,function()
    
    set_cg_param1f("draw_time",frame_time())
    --set_cg_param1f("draw_time",frame_time())
    set_perlin_texture()

    cg_texture_from_fullscreen_index('dtex',0)

    --gl_texture_barrier()

    glTextureBox(0,0,screen_w,screen_h,0,0,screen_w,screen_h)

  end)

end

--if not all_empires
--  draw_test=_nullop
--end



function on_topmost()
  print 'on_topmost'
  fullscreen_pane=split_screen_panes
  pane_alpha.jump_state(split_screen_panes)

  -- messy-edge case:
  --
  -- if we've just completed a battle from inside the tactical
  -- screen, we want the ai to immediately start up any additional
  -- battles that may be desirable given the new strategic
  -- situation at the star.
  --
  -- however, there's no easy way to trigger these from the
  -- battle complete callbacks -- as all those callbacks execute
  -- while the tactical screen is still topmost, and the attached
  -- battle state is still active.  thus, triggering the
  -- handle_potential_attacks_at callback from within one of
  -- those locations can lead to the AI trying to initialize one
  -- battle while another one is still in the process of being
  -- torn down -- a situation that leads to trouble.
  --
  -- what we want is for the AI's logic to start triggering as
  -- soon as the tactical screen is done being cleaned up. 
  -- we happen to know that the galaxy screens on_topmost
  -- callback will always be triggered at just that moment, so,
  -- we can leave a 'hint' for ourselves to trigger the AI
  -- actions from inside the end_battle callbacks, and then
  -- execute as needed here.
  if GalaxyMap.completed_attack_id
    local attack = all_attacks[completed_attack_id]
    GalaxyMap.completed_attack_id=nil
    if attack.star
      call_with_quiet_assignments(
        AI.handle_potential_attacks_at, attack.star)
    end
  end
end

function close_fullscreen_pane()
  --print(..stack_trace(1,6))
	local old_screen = fullscreen_pane
  old_screen?.on_close?()

	fullscreen_pane=split_screen_panes

  if old_screen?.call_on_close
    print(..fullscreen_pane)
    old_screen.call_on_close()
    print('called it',..fullscreen_pane)
    old_screen.call_on_close=nil
  end


  if (not FLAG.on_changed_galaxy) and (not FLAG.activating_unlisted_report) and is_report_key(selected_map_key)?.unlisted
    FLAG_call('soft_select', select_next_report_or_close)
  end
end


--pane_alpha=nil
pane_alpha = pane_alpha or AlphaState(4)
local function layout_scaled_draw()

  pane_alpha.update_with_active_key(fullscreen_pane)

  for pane,value in pane_alpha.pairs()
    SCOPE_frame_activation(  crossfade_alpha(value), pane.draw)    
  end

  dropdown_menu.draw_all_menus()
  fullscreen_pane.draw_cursor_fun?()?()

	--gui.cloud_spinner_overlay()

  --(fullscreen_pane!.draw)()
  draw_performance()
end

local function grey_alpha()
  glColor4d(1,1,1,.2)
--	local layout_w,layout_h in layout
  glBox(0,0,screen_w,screen_h)
end

local function fbo_draw()
  

  dropdown_menu.predraw_clears()
  glClearScreen(0,0,0,1)

  glColor4d(1,1,1,1)

  --console(..dims.bottom_bar())
  --update_dt()

  draw_starfield()

--	if fullscreen_pane==designer_pane
--		console(..fullscreen_pane?.draw)
--	else

  --glColor4d(1,1,1,1)

  --save_drawfun('test',layout.scaled_draw_function,layout_scaled_draw)
  fullscreen_pane=fullscreen_pane or split_screen_panes
  layout.scaled_draw_function( layout_scaled_draw)

--	save_drawfun('grey',grey_alpha)



end

local function attempt_draw()
  glColor4d(1,1,1,1)
  glNoTexture()
  glBlend | true
  
  local main_fbo = clean_buffer(1)
  call_with_fullscreen_fbo(main_fbo, SCOPE_call, .._ELEMENT_ENV, fbo_draw)
  draw_fullscreen_fbo(main_fbo,false,0,0,screen_w,screen_h)
  loading_fullscreen_texture_fbo = main_fbo
end

local function min_draw()
  draw_starfield()
  draw_top_bar()
end
  



local function recording_draw()

  clear_callbacks()
  check_order_callbacks()


  if true

    local ok, err=pcall(attempt_draw)
    if not ok
      add_exception(tostring | err)
      glClearScreen(0,0,0,0)		
    end

  else
    glClearScreen(0,0,0,0)	
    draw_performance()
  end
  draw_debugging_ui()
end

supports_qgl=true

function draw()

	if not using_qgl(_ENV)
		set_qgl_postpone_deletes(false)
		recording_draw()	
	else
		set_qgl_postpone_deletes(true)
		pcall(call_in_qgl_record_scope, recording_draw )
	end

  --dprint | check_ui_changed_callbacks()

end

function clear_hovers()
--	hover_star=nil
  --candidate_dest=nil
--	hover_fleet_id=nil

  if star_selection_mode
    star_selection_mode.candidate_star=nil
  end

end


function has_hover()
  return hover_star
end

function clear_true_selections()
  close_fullscreen_pane()
  close_midscreen_pane()

  --error("shouldn't have gotten here",4)

  if selected_map_key
    fallback_click_sound 'PANE_CLOSE'
  end

  selected_fleet_id=nil
  selected_map_key=nil
  selected_report=nil
  clear_transport()
end

function clear_selection()
  clear_hovers()
  clear_true_selections()
  close_midscreen_pane()
end

function undo_semantic()
  if shift_down()
    if ctrl_down()
      return jump_to_prev_year,10
    else
      return jump_to_prev_year
    end
  else
    return undo_order
  end
end

function redo_semantic()
  if shift_down()
    if ctrl_down()
      return jump_to_next_year,10
    else
      return jump_to_next_year
    end
  else
    return redo_order
  end
end



function order_hotkeys()

  if space_pressed()
    end_year_click()
  end

  
  if not gui.hide_undo_redo
    if alt_down()
      --sprint("alt")

      if left_pressed() 
       call_with_args( undo_semantic() )
      end

      if right_pressed() 
        call_with_args( redo_semantic() )  
      end
    end
  end
end


function open_main_menu()
    if fullscreen_pane~=split_screen_panes
      close_fullscreen_pane()
    end
    open_midscreen_pane(main_menu_pane)
end

function inhibit_hotkeys()
  return ship_status_pane.name_edit.is_active() or planet_name_edit.is_active() or trading_overlay.inhibit_hotkeys() or star_name_edit.is_active()
end

local on_click_hints = {}

local function on_no_drag()

  cleartable(on_click_hints)
  SCOPE_call(..on_click_hints, clear_selection)
  if not on_click_hints.played_sound
    local sound = on_click_hints.fallback_sound
    if sound
      Sounds![sound ]()
    end
  end
end

local function my_refresh(i,...)
  top_transport_key=nil

  if fullscreen_pane and rawget(fullscreen_pane,'refresh')
    --print('refreshing')
    return fullscreen_pane.refresh(i,...)
  end

  if not fullscreen_pane?.draw_cursor_fun?() 
    blue_cursor() 
  end

  clear_hovers()
  update_dt()
  update_xy()

  if funkey_pressed(10)
    open_main_menu()
  end

  --print(..fullscreen_pane)

  if scrolling
    scrolling(i,...)
    update_hover_alphas()
    return update_map_elements()
  end

  if star_selection_mode

    --print(..frame_time())

  		;(dispatch_zoom_overload or zoom_overload)?()
    return star_selection_refresh(i,...)
    
  end

  if planet_selection_mode
  		;(dispatch_zoom_overload or zoom_overload)?()
    return move_transports(i,...)
  end

  if starting_new_year
    select_next_report()
    starting_new_year=nil
  end

  --console 'refresh'
  update_map_elements(nelement_refresh(i,...))

  --[[
  if 	check_rclick_state()
    clear_true_selections()
  end
  --]]


  update_fleet_drag(i,...)



  if lmouse_pressed() and i==nil
    --print ('starting drag',frame_time())
    start_map_drag(on_no_drag)
--    clear_selection()
  end

  if fullscreen_pane==nil
    fullscreen_pane=split_screen_panes
  end

  if not inhibit_hotkeys()

    if (fullscreen_pane==split_screen_panes) or (fullscreen_pane==tech_pane)
      order_hotkeys()
    end

    fullscreen_pane.hotkeys?()

  end


  -- let us escape out of various sub-screens.
  -- the order of these checks is important, as it determines
  -- what gets closed first.
  if esc_pressed() then
    (fullscreen_pane.esc_handler or close_fullscreen_pane ) ()
  end

  --[=[

  -- test for '<'
  if char_pressed(string.char(188))
    select_next_report(selected_map_key,-1)

    --hide_cursor()
    --local rr=(skin.small_fonts and 195 or 240)+ (dec==1 and 20 or 30)
      
    --local ry=skin.bottom_bar.height+25
    --local rr=screen_w*.5 + (dec==1 and -30 or -20)
    --local rx = screen_w*.5+rr*dec

    local rx,ry=screen_w, bottom_bar.height

    set_cursor_pos( rx, ry)
    --cursor_fly_x,cursor_fly_y=	

  end

  -- test for '>'
  if char_pressed(string.char(190))
    select_next_report(selected_map_key)

    local rx,ry=screen_w, bottom_bar.height
    set_cursor_pos( rx, ry)
  end

  -]=]

  -- debugging hooks
  if ctrl_down()
  
    if char_pressed 'B'
      battle_list_pane.open()
    elseif char_pressed 'E' and options_pane.show_hidden_options()
      open_fullscreen_pane | event_trigger_pane
    elseif char_pressed 'T'
      open_fullscreen_pane | sound_test_pane
    elseif char_pressed 'U' and options.include_incomplete_features
      --push_screen 'StartScreen'
      --push_screen("TestScreen2")
      push_screen("UnpackScreen")
    elseif char_pressed 'R'
      --push_screen 'StartScreen'
      --push_screen("TestScreen2")
      --push_screen("MemoryTestScreen")
      open_fullscreen_pane | resolution_test_pane 
    elseif char_pressed 'Y'
      --console('Y')
      push_screen("PlanetTestScreen")
    elseif char_pressed 'O'
      open_fullscreen_pane | load_pane
    elseif char_pressed 'S'
      --open_fullscreen_pane | save_pane
    elseif alt_down() and char_pressed 'C'
      order.tech_cheat() 
    elseif char_pressed 'V' and enable_svens_special_hooks()

      open_fullscreen_pane | bug_pane
    end

  end

end

function post_draw(...)
  --	if fullscreen_pane==designer_pane
  --		return
  --	end

  if fullscreen_pane==split_screen_panes
    if (midscreen_pane~=diplo_pane) and (planet_selection_mode~=AlienPopDelivery) 
      Music.set_theme 'main'
    end
  end

  local screen_loop = fullscreen_pane?.environment_loop or 
    'SOLAR_GALAXY_SYSTEM_ENVIROMENTAL'
  
  if (fullscreen_pane==split_screen_panes) and 
    ( (midscreen_pane==diplo_pane) or  (planet_selection_mode==AlienPopDelivery) )
    screen_loop = 'DIPLOMACY_ENVIROMENTAL'
  end

  if ( (fullscreen_pane==split_screen_panes) and midscreen_pane==ship_status_pane and not ship_status_pane.object?.planet )
  else
    ensure_looping_sound_effect(screen_loop,screen_loop)
  end

  local ok, err=pcall(SCOPE_call, .._ELEMENT_ENV, layout.scaled_draw_function,my_refresh,... )

  if not ok
    add_exception | tostring | err
    --console('fallback refresh')
    nelement_refresh(...)
  end

  handle_hotkeys()
end

local function my_resize_callbacks()
  dropdown_menu.ensure_tables()
  layout_callback()
  layout.scaled_draw_function(execute_resize_callbacks)

  if all_empires regather_elements_if_needed() end

  reset_ship_layouts()
  layout.scaled_draw_function( reset_report_layouts) 
  --console(screen_w,screen_h)
end

function resize_callback()
  if fullscreen_pane==split_screen_panes
    Music.set_theme 'main'
  end

  SCOPE_call(.._ELEMENT_ENV, my_resize_callbacks)

end


function on_changed_galaxy()
  FLAG_call('on_changed_galaxy',|) function()
    GalaxyMap.completed_attack_id=nil
    planet_link_history.clear()

    ship_status_pane.object=nil
    split_screen_panes.update_alphas()
    update_tactical_colors()

    diplo_pane.on_changed_galaxy()
    clear_selection()

    clear_map_clicktimes()

    update_margins()
    local star = next(gui_player().stars)
    if star center_at_star(star) end

    reset_planet_name_cache()
    precache_planet_names()
    leftside_pane = planet_list_pane

    if selected_map_key
      instant_key_select | selected_map_key
    end
    tech_tree_pane.on_changed_galaxy()
		reset_territory_caches()
  end
end


if lua_load_complete 

  FLAG_call('script_reload',SCOPE_call,_ELEMENT_ENV=_ENV,select_map_key,selected_map_key)

  resize_callback()

end

--

--[[
do
  local whitelist=set("hover_star",
    "transport_landing","candidate_dest","current_map_key",
    "selected_map_key","invalid_transport_dest")

  local mt=getmetatable(_ENV) or {}
  local ni=mt.__newindex
  function mt.__newindex(t,k,v)

    if not whitelist[k] 
      console("set",k,stack_trace(2,6))
    end

    return (ni or rawset)(t,k,v)
  end

  setmetatable(_ENV,mt)

end
--]]